/*
 *  GPSdo - Software for GPSDO
 *  (c) Ad Sekeris 2019
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * File: adf4351_func.h - contains all encoder related functions
 */

// Many thanks to Antony M0IFA for the start

class ADF4351 {

  public:
    ADF4351();  // constructor
    
    void init();  // Initialize`
    
    void setFreq(unsigned long long RFout, unsigned long freqRes); // Set the output frequency
    
    void update(); // Update the registers

    void setRFOut(byte enabled);    // Enable or disable RF output

    void setRFPower(byte power);    // Set the RF output power
    
  private:

    void writeRegister(unsigned long data); // Write to a register on the device

    struct
    {
      int Integer; 
      unsigned int Fractional;
    } Register0;      //  R0
    
    struct
    {
      byte PhaseAdjust;
      byte Prescaler;
      unsigned int Phase;
      unsigned int Modulus;
    } Register1;      //R1
    
    struct
    {
      byte NoiseMode; // & Spur Mode
      byte MUXOut;
      byte ReferenceDoubler;
      byte ReferenceDivider;
      unsigned int RCounter;
      byte DoubleBuffer;
      byte ChargePumpCurrent;
      byte LDF;
      byte LDP;
      byte PDPolarity;
      byte PowerDown;
      byte CP;
      byte CounterReset;
    } Register2;      // R2
    
    struct
    {
      byte BandSelectClockMode;
      byte ABP;
      byte ChargeCancelation;
      byte CSR;
      byte ClockDividerMode;
      unsigned int ClockDivider;
    } Register3;      // R3
    
    struct
    {
      byte FBSelect;
      byte RFDivider;
      byte BandSelectDivider;
      byte VCOPoweredDown;
      byte MTLD;
      byte AuxOutputSelect;
      byte AuxOutputEnabled;
      byte AuxOutputPower;
      byte OutputEnabled;
      byte OutputPower;
    } Register4;      // R4
    
    struct
    {
      byte LDPinMode;
    } Register5;      // R5
    
    unsigned long r0;     // 32 bit registers
    unsigned long r1;
    unsigned long r2;
    unsigned long r3;
    unsigned long r4;
    unsigned long r5; 
    
}; 

ADF4351::ADF4351()  {          // constructor
}

void ADF4351::init() {
  pinMode(SS, OUTPUT);        // SPI enable
  digitalWrite(SS, HIGH);       // SS off

  SPI.begin(); // SPI, mode in update()
  
  // set register 0 default values
  Register0.Integer = 0; // set by setFreq()
  Register0.Fractional = 0; // set by setFreq()

  // Set register 1 default values
  Register1.PhaseAdjust = 0; // off
  Register1.Prescaler = 1; // 8/9
  Register1.Phase = 1; // 
  Register1.Modulus = 0;  // set by setFreq()
  
  // Set register 2 default values
  Register2.NoiseMode = 0; // b00,inc spur mode
  Register2.MUXOut = 0; // 3 state
  Register2.ReferenceDoubler = 0; // off
  Register2.ReferenceDivider = 0; // off
  Register2.RCounter = 4 ; // b0000000100, max 4095
  Register2.DoubleBuffer = 0; // off
  Register2.ChargePumpCurrent = 7 ; // b0111, 2.5mA
  Register2.LDF = 0; // FRACN
  Register2.LDP = 0; // 10ns lock detect window
  Register2.PDPolarity = 1; // +ve
  Register2.PowerDown = 0; // pwr dwn disabled
  Register2.CP = 0; // off
  Register2.CounterReset = 0; // off

  // Set register 3 default values
  Register3.BandSelectClockMode = 0; // low
  Register3.ABP = 0; // 6ns (FRAC-N)
  Register3.ChargeCancelation = 0; // off
  Register3.CSR = 0; // off
  Register3.ClockDividerMode = 0; // off
  Register3.ClockDivider = 75; // Band Sel clk max 125kHz

  // Set register 4 default values
  Register4.FBSelect = 1; // 1 from VCO, or 0 RFout
  Register4.RFDivider = 2; // b010
  Register4.BandSelectDivider = 20; // b00010100, auto set < 125kHz
  Register4.VCOPoweredDown = 0; // pwr up
  Register4.MTLD = 0; // off
  Register4.AuxOutputSelect = 0; // off
  Register4.AuxOutputEnabled = 0; // disable
  Register4.AuxOutputPower = 0; // -4dBm
  Register4.OutputEnabled = 0; // off
  Register4.OutputPower = 0; // -4dBm
}

void ADF4351::setFreq(unsigned long long RFout, unsigned long freqRes) { // Hz
  uint32_t fpfd;
  float fracf; 
  float OutDivider;

  // RFDIVIDER uint8_t
  Register4.RFDivider = 0;      // default
  OutDivider = 1.00;
  if (RFout >= 2200000000)      // Hz
  {
    Register4.RFDivider = 0;    // Divide by 2^0 = 1
    OutDivider = 1.00;
  }
  if (RFout < 2200000000)
  {
    Register4.RFDivider = 1;    // Divide by 2
    OutDivider = 2.00;
  }
  if (RFout < 1100000000)
  {
    Register4.RFDivider = 2;    // Divide by 4
    OutDivider = 4.00;
  }
  if (RFout < 550000000)
  {
    Register4.RFDivider = 3;    // Divide by 8
    OutDivider = 8.00;
  }
  if (RFout < 275000000)
  {
    Register4.RFDivider = 4;    // Divide by 16
    OutDivider = 16.00;
  }
  if (RFout < 137500000)
  {
    Register4.RFDivider = 5;    // Divide by 32
    OutDivider = 32.00;
  }
  if (RFout < 68750000)
  {
    Register4.RFDivider = 6;    // Divide by 64
    OutDivider = 64.00;
  }
  // FPFD uint32_t
  fpfd = xtal * (1 + Register2.ReferenceDoubler) / (Register2.RCounter * (1 + Register2.ReferenceDivider));

  // FRACF float
  fracf = (float)(RFout * OutDivider) / (float)fpfd;   // Float calc

  // INT uint16_t
  Register0.Integer = fracf;    // integer part of FRACF

  // MOD uint16_t
  Register1.Modulus = (float)((float)fpfd * (100000 /(float)freqRes) / 100000); // Phase freq det/channel spacing

  // FRAC uint16_t
  Register0.Fractional = round(((float)fracf - (float)Register0.Integer) * (float)Register1.Modulus);    // floating point calc, rounded

  // eliminate any common factors of 2, 3 or 5 to minimise spurs
  while (((Register1.Modulus % 2)==0) && ((Register0.Fractional % 2)==0)) {
    Register1.Modulus /= 2;
    Register0.Fractional /= 2;
  }
  while (((Register1.Modulus % 3)==0) && ((Register0.Fractional % 3)==0)) {
    Register1.Modulus /= 3;
    Register0.Fractional /= 3;
  }        
  while (((Register1.Modulus % 5)==0) && ((Register0.Fractional % 5)==0)) {
    Register1.Modulus /= 5;
    Register0.Fractional /= 5;
  }

#ifdef USB_DEBUG 
  char tx_data[] = "        "; 
  Serial.print("fpfd = ");
  //sprintf(tx_data, "%8lX", r0);
  Serial.println(fpfd);
  Serial.print("fracf = ");
  //sprintf(tx_data, "%8lX", r1);
  Serial.println(fracf);
  Serial.print("Register0.Integer = ");
  //sprintf(tx_data, "%8lX", r2);
  Serial.println(Register0.Integer);
  Serial.print("Register1.Modulus = ");
  //sprintf(tx_data, "%8lX", r3);
  Serial.println(Register1.Modulus);
  Serial.print("Register0.Fractional = ");
  //sprintf(tx_data, "%8lX", r4);
  Serial.println(Register0.Fractional);
  Serial.println();
#endif

}

void ADF4351::update()
{
  // Build register 0
  r0 = (uint32_t)(Register0.Integer) << 15 |
     (uint32_t)(Register0.Fractional) << 3 |
     0x00000000;
  
  // Build register 1
  r1 = (uint32_t)(Register1.PhaseAdjust) << 28 |
     (uint32_t)(Register1.Prescaler) << 27 |  
     (uint32_t)(Register1.Phase) << 15 |
     (uint32_t)(Register1.Modulus) << 3 |
     0x00000001;
  
  // Build register 2
  r2 = (uint32_t)(Register2.NoiseMode) << 29 |
     (uint32_t)(Register2.MUXOut) << 26 |
     (uint32_t)(Register2.ReferenceDoubler) << 25 |
     (uint32_t)(Register2.ReferenceDivider) << 24 |
     (uint32_t)(Register2.RCounter) << 14 |
     (uint32_t)(Register2.DoubleBuffer) << 13 |
     (uint32_t)(Register2.ChargePumpCurrent) << 9 |
     (uint32_t)(Register2.LDF) << 8 |
     (uint32_t)(Register2.LDP) << 7 |
     (uint32_t)(Register2.PDPolarity) << 6 |
     (uint32_t)(Register2.PowerDown) << 5 |
     (uint32_t)(Register2.CP) << 4 |
     (uint32_t)(Register2.CounterReset) << 3 |
     0x00000002;
  
  r3 = (uint32_t)(Register3.BandSelectClockMode) << 23 |
    (uint32_t)(Register3.ABP) << 22 |
    (uint32_t)(Register3.ChargeCancelation) << 21 |
    (uint32_t)(Register3.CSR) << 18 |
    (uint32_t)(Register3.ClockDividerMode) << 15 | 
    (uint32_t)(Register3.ClockDivider) << 3 |
    0x00000003;
  
  // Build register 4
  r4 = (uint32_t)(Register4.FBSelect) << 23 |
     (uint32_t)(Register4.RFDivider) << 20 |
     (uint32_t)(Register4.BandSelectDivider) << 12 |
     (uint32_t)(Register4.VCOPoweredDown) << 11 |
     (uint32_t)(Register4.MTLD) << 10 |
     (uint32_t)(Register4.AuxOutputSelect) << 9 | 
     (uint32_t)(Register4.AuxOutputEnabled) << 8 | 
     (uint32_t)(Register4.AuxOutputPower) << 6 | 
     (uint32_t)(Register4.OutputEnabled) << 5 | 
     (uint32_t)(Register4.OutputPower) << 3 | 
     0x00000004;

  // fix r5
  r5 = 0x00580005;
#ifdef USB_DEBUG 
  char tx_data[] = "        "; 
  Serial.print("R0 = ");
  sprintf(tx_data, "%8lX", r0);
  Serial.println(tx_data);
  Serial.print("R1 = ");
  sprintf(tx_data, "%8lX", r1);
  Serial.println(tx_data);
  Serial.print("R2 = ");
  sprintf(tx_data, "%8lX", r2);
  Serial.println(tx_data);
  Serial.print("R3 = ");
  sprintf(tx_data, "%8lX", r3);
  Serial.println(tx_data);
  Serial.print("R4 = ");
  sprintf(tx_data, "%8lX", r4);
  Serial.println(tx_data);
  Serial.print("R5 = ");
  sprintf(tx_data, "%8lX", r5);
  Serial.println(tx_data);
  Serial.println();
#endif
  // Write the registers  
  //noInterrupts();
//  SPI.beginTransaction(SPISettings(1000000, MSBFIRST, SPI_MODE0));
    SPI.beginTransaction(SPISettings(4000000, MSBFIRST, SPI_MODE0));

  writeRegister(r5);
  writeRegister(r4);
  writeRegister(r3);
  writeRegister(r2);
  writeRegister(r1);
  writeRegister(r0);

  SPI.endTransaction();
  //interrupts();
}

void ADF4351::writeRegister(uint32_t data)
{
  digitalWrite(SS, LOW);        // enable ADF4351 SPI
  for(int i = 3; i >= 0 ; i--)      // Writes the 32bit data bytes
  {
    SPI.transfer((data >> 8 * i) & 0xFF); // write byte
  }
  digitalWrite(SS, HIGH);       // latch in data, disable SPI
  delayMicroseconds(2000);      // 2ms delay of register writes
}

void ADF4351::setRFOut(uint8_t enabled)
{
  if (enabled == 0)
    Register4.OutputEnabled = 0;
  else
    Register4.OutputEnabled = 1;
}


void ADF4351::setRFPower(uint8_t power)
{
  // Disgard invalid power values
  if (power > 3)
    return;
  
  Register4.OutputPower = power;
}
